local lume = require('lib/lume')
local bresenham = require('lib/bresenham')
local Delaunay = require('lib/delaunay')
local Point = Delaunay.Point
local Map
do
  local _class_0
  local _base_0 = { }
  _base_0.__index = _base_0
  _class_0 = setmetatable({
    __init = function(self, width, height, grid)
      self.width = width
      self.height = height
      self.tiles = grid
    end,
    __base = _base_0,
    __name = "Map"
  }, {
    __index = _base_0,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  Map = _class_0
end
local rr
rr = function(x, y)
  return lume.round(lume.random(x, y))
end
local Grid
do
  local _class_0
  local _base_0 = {
    print = function(self)
      for y = 0, self.height do
        local line = ""
        for x = 0, self.width do
          local t = '0'
          if self.grid[x][y] == 1 then
            t = '.'
          end
          if self.grid[x][y] == 2 then
            t = ','
          end
          line = line .. t
        end
        print(line)
      end
    end,
    add_room = function(self, room)
      local hw = room.halfwidth
      local hh = room.halfheight
      for x = room.x - hw, room.x + hw do
        for y = room.y - hh, room.y + hh do
          self.grid[x][y] = 1
          self.room_grid[x][y] = #self.rooms
        end
      end
      return table.insert(self.rooms, room)
    end,
    test_room = function(self, room)
      local result = true
      local hw = room.halfwidth
      local hh = room.halfheight
      for x = room.x - hw - 1, room.x + hw + 1 do
        if result == false or x == 0 or x >= self.width then
          result = false
          break
        end
        for y = (room.y - hh - 1), (room.y + hh + 1) do
          if self.grid[x] == nil or self.grid[x][y] ~= 0 then
            result = false
            break
          end
        end
      end
      return result
    end,
    get_spawn_location = function(self)
      local tx = rr(1, self.width - 1)
      local ty = rr(1, self.height - 1)
      while self.items[tx][ty] ~= false or self.grid[tx][ty] ~= 1 do
        tx = rr(1, self.width - 1)
        ty = rr(1, self.height - 1)
      end
      self:set_item_spawn(tx, ty, true)
      return tx, ty
    end,
    set_item_spawn = function(self, x, y, item)
      self.items[x][y] = item
    end
  }
  _base_0.__index = _base_0
  _class_0 = setmetatable({
    __init = function(self, width, height, d)
      if d == nil then
        d = 0
      end
      self.width = width
      self.height = height
      self.grid = { }
      self.items = { }
      self.room_grid = { }
      for x = 0, width do
        self.grid[x] = { }
        self.items[x] = { }
        self.room_grid[x] = { }
        for y = 0, height do
          self.grid[x][y] = d
          self.items[x][y] = false
          self.room_grid[x][y] = -1
        end
      end
      self.rooms = { }
    end,
    __base = _base_0,
    __name = "Grid"
  }, {
    __index = _base_0,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  Grid = _class_0
end
local Room
do
  local _class_0
  local _base_0 = { }
  _base_0.__index = _base_0
  _class_0 = setmetatable({
    __init = function(self, x, y, width, height)
      self.x = x
      self.y = y
      self.halfwidth = width
      self.width = width * 2 + 1
      self.halfheight = height
      self.height = height * 2 + 1
    end,
    __base = _base_0,
    __name = "Room"
  }, {
    __index = _base_0,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  Room = _class_0
end
local setseed
setseed = function(seed)
  if seed == nil then
    seed = os.time()
  end
  math.randomseed(seed)
  return seed
end
local Desert
Desert = function(width, height, seed)
  setseed(seed)
  local grid = Grid(width, height, 1)
  local tilemap = {
    3,
    1
  }
  local entities = { }
  local tx, ty = grid:get_spawn_location()
  table.insert(entities, {
    x = tx,
    y = ty,
    type = 'pc'
  })
  for i = 1, 4 do
    tx, ty = grid:get_spawn_location()
    table.insert(entities, {
      x = tx,
      y = ty,
      type = 'mine_entrance',
      data = {
        interact_map = 'dungeon',
        seed = math.random() * 1000
      }
    })
  end
  for i = 1, 40 do
    tx, ty = grid:get_spawn_location()
    local data = {
      sprite_frame = {
        rr(2, 4),
        2
      }
    }
    table.insert(entities, {
      x = tx,
      y = ty,
      type = 'cactus',
      data = data
    })
  end
  for i = 1, 30 do
    tx, ty = grid:get_spawn_location()
    table.insert(entities, {
      x = tx,
      y = ty,
      type = 'snake'
    })
  end
  return grid, tilemap, entities
end
local Dungeon
Dungeon = function(width, height, seed)
  seed = setseed(seed)
  local grid = Grid(width, height)
  for i = 1, 30 do
    local test = 0
    local room = Room(rr(0, (width - 1) / 2) * 2 + 1, rr(0, (height - 1) / 2) * 2 + 1, 2, 2)
    while not grid:test_room(room) and test < 10 do
      room.x = rr(0, (width - 1) / 2) * 2 + 1
      room.y = rr(0, (height - 1) / 2) * 2 + 1
      test = test + 1
    end
    if test < 10 then
      local _ = nil
      grid:add_room(room)
    end
  end
  local tx, ty = 1, 1
  local bx, by = tx, ty
  local bread_crumbs = { }
  local maze
  maze = function(tx, ty)
    local tests = {
      {
        1,
        0
      },
      {
        -1,
        0
      },
      {
        0,
        1
      },
      {
        0,
        -1
      }
    }
    tests = lume.shuffle(tests)
    local moved = false
    for _index_0 = 1, #tests do
      local test = tests[_index_0]
      local _tx = tx + test[1] * 2
      local _ty = ty + test[2] * 2
      if _tx > 0 and _tx < width and _ty > 0 and _ty < height then
        if grid.grid[_tx][_ty] == 0 then
          grid.grid[tx + test[1]][ty + test[2]] = 1
          grid.grid[_tx][_ty] = 1
          grid.room_grid[tx + test[1]][ty + test[2]] = 0
          grid.room_grid[_tx][_ty] = 0
          maze(_tx, _ty)
        end
      end
    end
  end
  for x = 1, width, 2 do
    for y = 1, height, 2 do
      maze(x, y)
    end
  end
  local connectors = { }
  for x = 2, width - 2 do
    for y = 2, height - 2 do
      if grid.grid[x][y] == 0 then
        local a = grid.room_grid[x][y - 1]
        local b = grid.room_grid[x][y + 1]
        if a ~= b and a >= 0 and b >= 0 and (a == 0 or b == 0) then
          local room = b
          if a > 0 then
            room = a
          end
          if connectors[room] == nil then
            connectors[room] = { }
          end
          table.insert(connectors[room], {
            x,
            y
          })
        end
        a = grid.room_grid[x - 1][y]
        b = grid.room_grid[x + 1][y]
        if a ~= b and (a > 0 or b > 0) and (a == 0 or b == 0) then
          local room = b
          if a > 0 then
            room = a
          end
          if connectors[room] == nil then
            connectors[room] = { }
          end
          table.insert(connectors[room], {
            x,
            y
          })
        end
      end
    end
  end
  for room, conns in pairs(connectors) do
    local _ = lume.randomchoice(conns)
    local x, y = _[1], _[2]
    grid.grid[x][y] = 1
  end
  local points = { }
  local _list_0 = grid.rooms
  for _index_0 = 1, #_list_0 do
    local room = _list_0[_index_0]
    table.insert(points, Point(room.x, room.y))
  end
  local tilemap = {
    3,
    7
  }
  local start_room = lume.randomchoice(grid.rooms)
  local entities = { }
  tx = start_room.x
  ty = start_room.y
  while grid.grid[tx][ty] == 1 do
    ty = ty - 1
  end
  ty = ty + 1
  table.insert(entities, {
    x = tx,
    y = ty - 0.5,
    type = 'mine_entrance',
    data = {
      interact_map = 'desert'
    }
  })
  table.insert(entities, {
    x = tx,
    y = ty,
    type = 'pc'
  })
  local _list_1 = grid.rooms
  for _index_0 = 1, #_list_1 do
    local room = _list_1[_index_0]
    table.insert(entities, {
      x = rr(room.x - room.halfwidth, room.x + room.halfwidth),
      y = rr(room.y - room.halfheight, room.y + room.halfheight),
      type = 'crate'
    })
  end
  for i = 1, 30 do
    tx, ty = grid:get_spawn_location()
    table.insert(entities, {
      x = tx,
      y = ty,
      type = 'snake'
    })
  end
  return grid, tilemap, entities
end
return {
  Desert = Desert,
  Dungeon = Dungeon
}
