local deep_copy, deep_extend, pprint
do
  local _obj_0 = require("utils")
  deep_copy, deep_extend, pprint = _obj_0.deep_copy, _obj_0.deep_extend, _obj_0.pprint
end
local TYPES = {
  default = {
    tx = 0,
    ty = 0,
    rot = 0,
    name = 'entity'
  },
  chara = {
    _inherit = 'default',
    sprite_image = "sprites",
    sprite_frame = {
      1,
      3
    },
    sprite_color = {
      255,
      255,
      255
    },
    vx = 0,
    vy = 0,
    ax = 0,
    ay = 0,
    acc = 512
  },
  mine_entrance = {
    _inherit = 'default',
    sprite_image = "sprites",
    sprite_frame = {
      8,
      2
    },
    sprite_color = {
      255,
      255,
      255
    },
    script = 'mineentrance',
    interact = {
      label = "Leave Mine"
    }
  },
  cactus = {
    _inherit = 'default',
    sprite_image = "sprites",
    sprite_frame = {
      2,
      2
    },
    sprite_color = {
      255,
      255,
      255
    }
  },
  crate = {
    _inherit = 'default',
    sprite_image = "sprites",
    sprite_frame = {
      10,
      2
    },
    sprite_color = {
      255,
      255,
      255
    },
    interact = {
      label = "Open Crate"
    }
  },
  box = {
    _inherit = 'default',
    sprite_image = "sprites",
    sprite_frame = {
      10,
      2
    },
    sprite_color = {
      255,
      255,
      255
    },
    interact = {
      label = "Open Box"
    }
  },
  ammo = {
    _inherit = 'default',
    sprite_image = "sprites",
    sprite_frame = {
      12,
      3
    },
    sprite_color = {
      255,
      255,
      255
    },
    sprite_scale = 1,
    sprite_oy = 12,
    trigger = true,
    droppable_attr = 'ammo',
    droppable_ammount = 6
  },
  red_potion = {
    _inherit = 'default',
    sprite_image = "sprites",
    sprite_frame = {
      11,
      2
    },
    sprite_color = {
      255,
      255,
      255
    },
    sprite_scale = 1,
    sprite_oy = 12,
    trigger = true,
    droppable_attr = 'health',
    droppable_ammount = 6
  },
  green_potion = {
    _inherit = 'default',
    sprite_image = "sprites",
    sprite_frame = {
      12,
      2
    },
    sprite_color = {
      255,
      255,
      255
    },
    sprite_scale = 1,
    sprite_oy = 12,
    trigger = true,
    droppable_attr = 'health',
    droppable_ammount = 6
  },
  yellow_potion = {
    _inherit = 'default',
    sprite_image = "sprites",
    sprite_frame = {
      13,
      2
    },
    sprite_color = {
      255,
      255,
      255
    },
    sprite_scale = 1,
    sprite_oy = 12,
    trigger = true,
    droppable_attr = 'health',
    droppable_ammount = 6
  },
  pc = {
    _inherit = 'chara',
    player = true,
    acc = 1024,
    health = 20,
    health_max = 20,
    name = "pc",
    melee_attack = 0,
    melee_damage = 5,
    melee_range = 24,
    anim_right = true,
    anim_track = 'idle',
    anim_tracks = {
      idle = {
        1,
        3
      },
      walk = {
        '2-3',
        3
      },
      melee = {
        '1-4',
        4
      },
      range = {
        '1-4',
        5
      }
    },
    ammo_max = 36,
    ammo = 36
  },
  enemy = {
    _inherit = 'default',
    sprite_image = "sprites",
    sprite_frame = {
      '2-3',
      3
    },
    sprite_color = {
      255,
      255,
      255
    },
    vx = 0,
    vy = 0,
    ax = 0,
    ay = 0,
    acc = 512,
    health = 10,
    anim_right = true,
    anim_track = 'idle'
  },
  snake = {
    _inherit = 'enemy',
    health = 6,
    sprite_frame = {
      9,
      1
    },
    name = 'snake',
    script = 'snake',
    acc = 128,
    speedMax = 12,
    melee_attack = 0,
    melee_damage = 1,
    melee_range = 12,
    sprite_oy = 12,
    anim_track = 'idle',
    anim_tracks = {
      idle = {
        '10-11',
        1
      },
      walk = {
        '12-15',
        1
      },
      attack = {
        '16-19',
        1
      }
    },
    width = 12,
    height = 12
  },
  fox = {
    name = 'fox',
    _inherit = 'enemy',
    health = 6,
    sprite_frame = {
      9,
      2
    }
  },
  bullet = {
    _inherit = 'default',
    name = 'bullet',
    vx = 0,
    vy = 0,
    ax = 0,
    ay = 0,
    acc = 2048,
    sprite_image = "sprites",
    sprite_frame = {
      6,
      2
    },
    sprite_color = {
      255,
      255,
      255
    },
    sprite_oy = 12,
    sprite_ox = 12,
    speedMax = 2048,
    solid = false,
    drag = 1,
    script = 'bullet'
  }
}
local EntityFactory
EntityFactory = function(data, defaults)
  if data == nil then
    data = { }
  end
  local default = {
    tx = 0,
    ty = 0,
    rot = 0
  }
  if defaults ~= nil then
    local tmp = TYPES[defaults]
    local inherits = {
      tmp
    }
    while tmp._inherit ~= nil do
      tmp = TYPES[tmp._inherit]
      table.insert(inherits, 1, tmp)
    end
    for _index_0 = 1, #inherits do
      local d = inherits[_index_0]
      deep_extend(default, d)
    end
  end
  local e = default
  deep_extend(e, data)
  return e
end
return {
  EntityFactory = EntityFactory
}
