local talkback = require('lib/talkback')
local lume = require('lib/lume')
local bresenham = require('lib/bresenham')
local Delaunay = require('lib/delaunay')
local Point = Delaunay.Point
local fov = require("lib/rsfov")
local bind_methods
bind_methods = require("lib/moon").bind_methods
local g = love.graphics
local Map
do
  local _class_0
  local _base_0 = {
    pov = function(self, x, y)
      for x = 1, self.width do
        for y = 1, self.height do
          if self.viz[x][y] == 1 then
            self.viz[x][y] = 2
          end
        end
      end
      local tx = math.floor(x / self.tilesize)
      local ty = math.floor(y / self.tilesize)
      fov(tx + 1, ty + 1, 7, self.methods.is_opaque, self.methods.set_visible)
      self.tilebatchdirty = true
    end,
    resize_screen = function(self, w, h)
      self.screenwidth = w
      self.screenheight = h
      self.tileviewwidth = math.ceil(self.screenwidth / self.tilesize)
      self.tileviewheight = math.ceil(self.screenheight / self.tilesize)
      self.tilebatch = love.graphics.newSpriteBatch(self.tile_image, self.tileviewwidth * self.tileviewheight)
      self.tilebatchdirty = true
      return print(self.tileviewwidth, self.tileviewheight)
    end,
    make_quads = function(self, image, width, height)
      local quads = { }
      local img_width = image:getWidth()
      local img_height = image:getHeight()
      local cols = img_width / width
      local rows = img_height / height
      for y = 0, rows - 1 do
        for x = 0, cols - 1 do
          table.insert(quads, g.newQuad(x * width, y * height, width, height, img_width, img_height))
        end
      end
      return quads
    end,
    load_sprite = function(self, image, width, height)
      local img = g.newImage('assets/' .. image .. '.png')
      img:setFilter('linear', 'nearest')
      local quads = self:make_quads(img, width, height)
      return img, quads
    end,
    update_tile_batch = function(self, camera)
      self.tilebatch:clear()
      local vieww = self.screenwidth / camera.scale
      local viewh = self.screenheight / camera.scale
      local startx = math.max(1, math.floor((camera.x - vieww / 2) / self.tilesize))
      local starty = math.max(1, math.floor((camera.y - viewh / 2) / self.tilesize))
      local endx = math.min(math.ceil(vieww / self.tilesize) + startx + 1, self.width)
      local endy = math.min(math.ceil(viewh / self.tilesize) + starty + 1, self.height)
      for row = startx, endx do
        for col = starty, endy do
          local x = row - 1
          local y = col - 1
          if self.viz[row][col] > 0 then
            if self.viz[row][col] == 1 then
              self.tilebatch:setColor(255, 255, 255)
            end
            if self.viz[row][col] == 2 then
              self.tilebatch:setColor(128, 128, 128)
            end
            self.tilebatch:add(self.tile_quads[self.tiles[row][col]], math.floor(x * self.tilesize), math.floor(y * self.tilesize))
          end
        end
      end
      self.tilebatchdirty = false
    end,
    is_visible = function(self, x, y)
      local tx = math.floor(x / self.tilesize)
      local ty = math.floor(y / self.tilesize)
      if tx < 0 or tx > self.width or ty < 0 or ty > self.height then
        return false
      end
      return self.viz[tx + 1][ty + 1]
    end,
    get_tile = function(self, x, y)
      local tx = math.floor(x / self.tilesize)
      local ty = math.floor(y / self.tilesize)
      return self.tiles[tx + 1][ty + 1]
    end,
    draw = function(self, camera)
      if self.oldcamera == nil then
        self.oldcamera = { }
        self.oldcamera.x = camera.x
        self.oldcamera.y = camera.y
      end
      g.setColor(255, 255, 255, 255)
      if self.oldcamera.x ~= camera.x or self.oldcamera.y ~= camera.y then
        self.tilebatchdirty = true
      end
      if self.tilebatchdirty then
        self:update_tile_batch(camera)
      end
      self.oldcamera.x = camera.x
      self.oldcamera.y = camera.y
      return love.graphics.draw(self.tilebatch)
    end,
    set_visible = function(self, tx, ty)
      if tx > 0 and tx < self.width and ty > 0 and ty < self.height then
        self.viz[tx][ty] = 1
      end
    end,
    is_opaque = function(self, x, y)
      if x <= 0 or x > self.width or y <= 0 or y >= self.height then
        return false
      end
      return self.tiles[x][y] ~= 3
    end,
    is_solid = function(self, tile, y)
      if y ~= nil then
        tile = self:get_tile(tile, y)
      end
      return tile == 3
    end,
    is_blocker = function(self, tile)
      return tile == 1
    end
  }
  _base_0.__index = _base_0
  _class_0 = setmetatable({
    __init = function(self, width, height, grid, tilemap)
      self.methods = bind_methods(self)
      self.events = talkback.new()
      self.scale = 3
      self.width = (width * self.scale) + 1
      self.height = (height * self.scale) + 1
      self.tiles = { }
      self.viz = { }
      self.tilesize = 16
      self.screenwidth = g.getWidth()
      self.screenheight = g.getHeight()
      self.tile_image, self.tile_quads = self:load_sprite('tiles/tiles', self.tilesize, self.tilesize)
      for row = 0, width do
        for i = 1, self.scale do
          self.viz[(row * self.scale) + i] = { }
          self.tiles[(row * self.scale) + i] = { }
        end
        for col = 0, height do
          if tilemap ~= nil then
            for x = 0, 2 do
              for y = 0, 2 do
                self.viz[(row * self.scale) + x + 1][(col * self.scale) + y + 1] = 0
                self.tiles[(row * self.scale) + x + 1][(col * self.scale) + y + 1] = tilemap[grid[row][col] + 1]
              end
            end
          else
            self.tiles[row][col] = grid[row - 1][col - 1] + 1
          end
        end
      end
      return self:resize_screen(self.screenwidth, self.screenheight)
    end,
    __base = _base_0,
    __name = "Map"
  }, {
    __index = _base_0,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  Map = _class_0
end
local map
map = function(width, height, seed)
  if seed == nil then
    seed = os.time()
  end
  math.randomseed(seed)
  local grid = { }
  for x = 1, width do
    grid[x] = { }
    for y = 1, height do
      grid[x][y] = 0
    end
  end
  local points = { }
  for i = 1, 5 do
    local room = {
      w = lume.round(lume.random(1, 7) * 2 + 1),
      h = lume.round(lume.random(1, 3) * 2 + 1)
    }
    local hw = room.w / 2
    local hh = room.h / 2
    local done = false
    while not done do
      room.x = lume.round(lume.random(1, width - 1))
      room.y = lume.round(lume.random(1, height - 1))
      if room.x - hw > 0 and room.x + hw < width and room.y - hh > 0 and room.y + hh < height then
        done = true
        for x = math.floor(room.x - hw) - 1, math.ceil(room.x + hw) + 1 do
          if done == false or x <= 0 or x >= height then
            done = false
            break
          end
          for y = math.floor(room.y - hh) - 1, math.ceil(room.y + hh) + 1 do
            if grid[x][y] == 1 or grid[x][y] == nil then
              done = false
              break
            end
          end
        end
      end
      if done then
        table.insert(points, Point(room.x, room.y))
      end
    end
    for x = math.floor(room.x - hw), math.ceil(room.x + hw) do
      for y = math.floor(room.y - hh), math.ceil(room.y + hh) do
        grid[x][y] = 1
      end
    end
  end
  local triangles = Delaunay.triangulate(unpack(points))
  local draw
  draw = function(x, y)
    grid[x][y] = 1
    return true
  end
  for i, triangle in ipairs(triangles) do
    local _list_0 = {
      triangle.e1,
      triangle.e2,
      triangle.e3
    }
    for _index_0 = 1, #_list_0 do
      local edge = _list_0[_index_0]
      local a = edge.p1
      local b = edge.p2
      local x, y = edge:getMidPoint()
      local mid = Point(lume.round(x), lume.round(y))
      local horz = math.random() > 0.5
      if horz then
        bresenham.los(b.x, b.y, b.x, mid.y, draw)
        bresenham.los(a.x, mid.y, b.x, mid.y, draw)
        bresenham.los(a.x, mid.y, a.x, a.y, draw)
      else
        bresenham.los(mid.x, a.y, a.x, a.y, draw)
        bresenham.los(mid.x, a.y, mid.x, b.y, draw)
        bresenham.los(b.x, b.y, mid.x, b.y, draw)
      end
    end
  end
  for y = 1, height do
    local line = ""
    for x = 1, width do
      local t = '0'
      if grid[x][y] == 1 then
        t = '.'
      end
      line = line .. t
    end
  end
  return grid
end
return {
  map = map,
  Map = Map
}
