local flux = require('lib/flux')
local tiny = require('lib/tiny')
local lume = require('lib/lume')
local bump = require('lib/bump')
local gamera = require('lib/gamera')
local talkback = require('lib/talkback')
local Vector = require('lib/vector-light')
local EntityFactory
EntityFactory = require("entity").EntityFactory
local deep_copy, deep_extend, q
do
  local _obj_0 = require("utils")
  deep_copy, deep_extend, q = _obj_0.deep_copy, _obj_0.deep_extend, _obj_0.q
end
local bind_methods
bind_methods = require("lib/moon").bind_methods
local GameState
GameState = require("gamestate").GameState
local GameMode
GameMode = require("gamemode").GameMode
local Controls
Controls = require("input").Controls
local Map, map
do
  local _obj_0 = require("map")
  Map, map = _obj_0.Map, _obj_0.map
end
local Desert, Dungeon
do
  local _obj_0 = require("dungeon")
  Desert, Dungeon = _obj_0.Desert, _obj_0.Dungeon
end
local TriggerSystem, InteractSystem, AnimSystem, RenderSystem, MoveSystem, PlayerSystem, MeleeSystem, LivingSystem, ScriptSystem
do
  local _obj_0 = require("systems")
  TriggerSystem, InteractSystem, AnimSystem, RenderSystem, MoveSystem, PlayerSystem, MeleeSystem, LivingSystem, ScriptSystem = _obj_0.TriggerSystem, _obj_0.InteractSystem, _obj_0.AnimSystem, _obj_0.RenderSystem, _obj_0.MoveSystem, _obj_0.PlayerSystem, _obj_0.MeleeSystem, _obj_0.LivingSystem, _obj_0.ScriptSystem
end
local g = love.graphics
local GameOverMode
do
  local _class_0
  local _parent_0 = GameMode
  local _base_0 = {
    update = function(self, dt)
      return self.state.world:update(dt)
    end,
    start = function(self)
      self.state.mode = "gameover"
      self.anim = {
        text_alpha = 0,
        bg_alpha = 0
      }
      return self.state.flux:to(self.anim, 1, {
        bg_alpha = 72
      }):after(self.anim, 1, {
        text_alpha = 255
      })
    end,
    draw = function(self)
      self.state:draw_game()
      g.setColor(255, 0, 0, self.anim.bg_alpha)
      g.rectangle("fill", 0, 0, self.state.engine.width, self.state.engine.height)
      g.setColor(255, 255, 255, self.anim.text_alpha)
      local width = 500
      return g.printf("Game Over", self.state.engine.width / 2 - width / 2, self.state.engine.height / 2, width, "center")
    end
  }
  _base_0.__index = _base_0
  setmetatable(_base_0, _parent_0.__base)
  _class_0 = setmetatable({
    __init = function(self, state)
      _class_0.__parent.__init(self, state)
      self.anim = {
        text_alpha = 0,
        bg_alpha = 0
      }
      return state.events:listen("gameover", bind_methods(self).start)
    end,
    __base = _base_0,
    __name = "GameOverMode",
    __parent = _parent_0
  }, {
    __index = function(cls, name)
      local val = rawget(_base_0, name)
      if val == nil then
        local parent = rawget(cls, "__parent")
        if parent then
          return parent[name]
        end
      else
        return val
      end
    end,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  if _parent_0.__inherited then
    _parent_0.__inherited(_parent_0, _class_0)
  end
  GameOverMode = _class_0
end
local PauseMode
do
  local _class_0
  local _parent_0 = GameMode
  local _base_0 = {
    update = function(self)
      if Controls.Primary:pressed() then
        self.state.mode = 'play'
      end
    end,
    draw = function(self)
      self.state:draw_bg()
      g.setColor(0, 0, 0, 128)
      g.rectangle("fill", 0, 0, self.state.engine.width, self.state.engine.height)
      g.setColor(255, 255, 255, 255)
      local width = 500
      return g.printf("Paused - Press Spacebar to unpause.", self.state.engine.width / 2 - width / 2, self.state.engine.height / 2, width, "center")
    end
  }
  _base_0.__index = _base_0
  setmetatable(_base_0, _parent_0.__base)
  _class_0 = setmetatable({
    __init = function(self, ...)
      return _class_0.__parent.__init(self, ...)
    end,
    __base = _base_0,
    __name = "PauseMode",
    __parent = _parent_0
  }, {
    __index = function(cls, name)
      local val = rawget(_base_0, name)
      if val == nil then
        local parent = rawget(cls, "__parent")
        if parent then
          return parent[name]
        end
      else
        return val
      end
    end,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  if _parent_0.__inherited then
    _parent_0.__inherited(_parent_0, _class_0)
  end
  PauseMode = _class_0
end
local PlayMode
do
  local _class_0
  local _parent_0 = GameMode
  local _base_0 = {
    update = function(self, dt)
      return self.state.world:update(dt)
    end,
    draw = function(self)
      return self.state:draw_game()
    end
  }
  _base_0.__index = _base_0
  setmetatable(_base_0, _parent_0.__base)
  _class_0 = setmetatable({
    __init = function(self, ...)
      return _class_0.__parent.__init(self, ...)
    end,
    __base = _base_0,
    __name = "PlayMode",
    __parent = _parent_0
  }, {
    __index = function(cls, name)
      local val = rawget(_base_0, name)
      if val == nil then
        local parent = rawget(cls, "__parent")
        if parent then
          return parent[name]
        end
      else
        return val
      end
    end,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  if _parent_0.__inherited then
    _parent_0.__inherited(_parent_0, _class_0)
  end
  PlayMode = _class_0
end
local PlayState
do
  local _class_0
  local _parent_0 = GameState
  local _base_0 = {
    init = function(self)
      self.events = talkback.new()
      self.modes = {
        play = PlayMode(self),
        pause = PauseMode(self),
        gameover = GameOverMode(self)
      }
      self.mode = 'play'
      self.ready = false
      self.map = nil
      self.load_progress = 1
      self._nextid = 0
      self.flux = flux.group()
      self.events:listen("player created", self.methods.game_start)
      return self.events:listen("player killed", self.methods.game_over)
    end,
    draw = function(self)
      return self.modes[self.mode]:draw()
    end,
    draw_game = function(self)
      if self.ready and self.load_progress < 1 then
        self:draw_bg()
        self.cam:draw(function(l, t, w, h)
          self.map:draw(self.cam)
          if self.renderer.entities then
            return self.renderer:draw()
          end
        end)
        self:draw_ui()
        g.setColor(0, 0, 0, 255 * self.load_progress)
        g.rectangle("fill", 0, 0, self.engine.width, self.engine.height)
        return g.setColor(255, 255, 255, 255)
      else
        g.setColor(0, 0, 0, 255)
        g.rectangle("fill", 0, 0, self.engine.width, self.engine.height)
        return g.setColor(255, 255, 255, 255)
      end
    end,
    draw_ui = function(self)
      if self.pc ~= nil and self.ready then
        g.push()
        g.translate(self.engine.width - 260, 0)
        g.printf(self.pc.ammo .. "/" .. self.pc.ammo_max, 0, 30, 120, "center")
        g.pop()
        g.push()
        g.translate(self.engine.width - 120, 0)
        g.setColor(255, 255, 255)
        g.rectangle("line", 10, 10, 100, 40)
        g.setColor(255, 0, 0)
        g.rectangle("fill", 10, 10, (self.pc.health / self.pc.health_max) * 100, 40)
        g.setColor(255, 255, 255)
        g.pop()
        if self.pc.interact_label ~= nil then
          g.push()
          g.translate(self.engine.width / 2 - 200, self.engine.height * 0.8)
          g.printf(self.pc.interact_label, 0, 0, 400, "center")
          return g.pop()
        end
      end
    end,
    draw_bg = function(self)
      g.setColor(0, 0, 0)
      g.rectangle("fill", 0, 0, self.engine.width, self.engine.height)
      return g.setColor(255, 255, 255)
    end,
    resize = function(self, w, h)
      if self.map ~= nil then
        return self.map:resize_screen(w, h)
      end
    end,
    update = function(self, dt)
      self.flux:update(dt)
      return self.modes[self.mode]:_update(dt)
    end,
    load = function(self, level, seed)
      self:reset()
      self:init_game()
      if level == nil then
        return self:new_game()
      else
        return self:load_game(level, seed)
      end
    end,
    create_entity = function(self, data, tmpl)
      local e = EntityFactory(data, tmpl)
      e.id = self._nextid
      self._nextid = self._nextid + 1
      e.events = talkback.new()
      e.events:listen("kill", function()
        return self.methods.remove_entity(e)
      end)
      return self.world:add(e)
    end,
    remove_entity = function(self, e)
      self.world:remove(e)
      if self.physics:hasItem(e) then
        return self.physics:remove(e)
      end
    end,
    search = function(self, tx, ty, radius, filter)
      local _results = self.physics:queryRect(tx - radius, ty - radius, radius * 2, radius * 2, filter)
      local results = { }
      for _index_0 = 1, #_results do
        local e = _results[_index_0]
        local length = Vector.dist(tx, ty, e.tx, e.ty)
        if length <= radius then
          table.insert(results, e)
        end
      end
      return results
    end,
    init_game = function(self)
      self.world = tiny.world()
      self.physics = bump.newWorld()
      self.cam = gamera.new(0, 0, 1000, 1000)
      self.cam.scale = 2
      self.cam:setPosition(500, 500)
      self.renderer = RenderSystem(self)
      self.world:add(self.renderer)
      self.world:add(MoveSystem(self))
      self.world:add(PlayerSystem(self))
      self.world:add(MeleeSystem(self))
      self.world:add(LivingSystem(self))
      self.world:add(AnimSystem(self))
      self.world:add(ScriptSystem(self))
      self.world:add(InteractSystem(self))
      return self.world:add(TriggerSystem(self))
    end,
    get_spawn_location = function(self)
      local tx = lume.round(lume.random(1, self.map.width))
      local ty = lume.round(lume.random(1, self.map.height))
      while self.map:is_solid(self.map.tiles[tx][ty]) do
        tx = lume.round(lume.random(1, self.map.width))
        ty = lume.round(lume.random(1, self.map.height))
      end
      return tx * self.map.tilesize, ty * self.map.tilesize
    end,
    game_start = function(self, pc)
      self.pc = pc
      if self.playthrough.pc ~= nil then
        self.pc.health = self.playthrough.pc.health
        self.playthrough.pc = nil
      end
    end,
    game_over = function(self)
      self.events:say("gameover")
      return self:Timer(3, function()
        return self.engine:load_state("mainmenu")
      end)
    end,
    new_game = function(self)
      self.playthrough = {
        seed = os.time()
      }
      local w = 11
      local h = 11
      local grid, tilemap, entities = Desert(w, h, self.playthrough.seed)
      return self:create_map(grid, tilemap, entities)
    end,
    load_game = function(self, typ, seed)
      seed = seed or 0
      local map_table = {
        dungeon = Dungeon,
        desert = Desert
      }
      local w = 27
      local h = 21
      if typ == desert then
        w = 11
        h = 11
      end
      local grid, tilemap, entities = map_table[typ](w, h, self.playthrough.seed + seed)
      return self:create_map(grid, tilemap, entities)
    end,
    create_map = function(self, grid, tilemap, entities)
      self.map = Map(grid.width, grid.height, grid.grid, tilemap)
      self.cam:setWorld(0, 0, self.map.width * self.map.tilesize, self.map.height * self.map.tilesize)
      for x = 1, self.map.width do
        for y = 1, self.map.height do
          local tile = self.map.tiles[x][y]
          if self.map:is_solid(tile) then
            self.physics:add({
              name = "tile" .. x .. "x" .. y,
              tile = tile,
              events = self.map.events
            }, (x - 1) * self.map.tilesize, (y - 1) * self.map.tilesize, self.map.tilesize, self.map.tilesize)
          end
        end
      end
      for _index_0 = 1, #entities do
        local e = entities[_index_0]
        e.data = e.data or { }
        deep_extend(e.data, {
          tx = (e.x + 0.5) * self.map.tilesize * self.map.scale,
          ty = (e.y + 0.5) * self.map.tilesize * self.map.scale
        })
        self:create_entity(e.data, e.type)
      end
      self.ready = true
      self.mode = 'play'
      return self.flux:to(self, 1, {
        load_progress = 0
      })
    end
  }
  _base_0.__index = _base_0
  setmetatable(_base_0, _parent_0.__base)
  _class_0 = setmetatable({
    __init = function(self, ...)
      return _class_0.__parent.__init(self, ...)
    end,
    __base = _base_0,
    __name = "PlayState",
    __parent = _parent_0
  }, {
    __index = function(cls, name)
      local val = rawget(_base_0, name)
      if val == nil then
        local parent = rawget(cls, "__parent")
        if parent then
          return parent[name]
        end
      else
        return val
      end
    end,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  if _parent_0.__inherited then
    _parent_0.__inherited(_parent_0, _class_0)
  end
  PlayState = _class_0
end
return {
  PlayState = PlayState
}
