local fov = require("lib/rsfov")
local flux = require('lib/flux')
local tiny = require('lib/tiny')
local lume = require('lib/lume')
local bump = require('lib/bump')
local anim8 = require('lib/anim8')
local gamera = require('lib/gamera')
local talkback = require('lib/talkback')
local Vector = require('lib/vector-light')
local bind_methods
bind_methods = require("lib/moon").bind_methods
local GameMode
GameMode = require("gamemode").GameMode
local Controls
Controls = require("input").Controls
local deep_copy, pprint
do
  local _obj_0 = require("utils")
  deep_copy, pprint = _obj_0.deep_copy, _obj_0.pprint
end
local g = love.graphics
local SPRITE_CACHE = { }
local SCRIPT_CACHE = { }
local TriggerSystem
TriggerSystem = function(state)
  local sys = tiny.processingSystem()
  sys.filter = tiny.requireAll("trigger")
  sys.onAdd = function(self, e)
    e.width, e.height = 12, 12
    e.solid = false
    if not state.physics:hasItem(e) then
      state.physics:add(e, e.tx - e.width / 2, e.ty - e.height / 2, e.width, e.height)
    end
    return e.events:listen("collide", function(o)
      if o.name == "pc" then
        e.events:say("kill")
        if e.droppable_attr then
          o[e.droppable_attr] = math.min(o[e.droppable_attr] + e.droppable_ammount, o[e.droppable_attr .. '_max'])
        end
      end
    end)
  end
  sys.process = function(self, e, dt)
    return nil
  end
  return sys
end
local InteractSystem
InteractSystem = function(state)
  local sys = tiny.processingSystem()
  sys.filter = tiny.requireAll("interact")
  sys.onAdd = function(self, e)
    e.width, e.height = 6, 6
    e.highlight_radius = 24
    if not state.physics:hasItem(e) then
      return state.physics:add(e, e.tx - e.width / 2, e.ty - e.height / 2, e.width, e.height)
    end
  end
  sys.preWrap = function(self, dt)
    local _list_0 = self.entities
    for _index_0 = 1, #_list_0 do
      local e = _list_0[_index_0]
      e.highlight = nil
    end
  end
  sys.process = function(self, e, dt)
    return nil
  end
  return sys
end
local ScriptSystem
ScriptSystem = function(state)
  local sys = tiny.processingSystem()
  sys.filter = tiny.requireAll("script")
  sys.onAdd = function(self, e)
    if SCRIPT_CACHE[e.script] == nil then
      local cmpt = require("scripts/" .. e.script)
      SCRIPT_CACHE[e.script] = cmpt(state)
    end
    return SCRIPT_CACHE[e.script]:onAdd(e)
  end
  sys.process = function(self, e, dt)
    return SCRIPT_CACHE[e.script]:update(e, dt)
  end
  return sys
end
local LivingSystem
LivingSystem = function(state)
  local sys = tiny.processingSystem()
  sys.filter = tiny.requireAll("health")
  sys.onAdd = function(self, e)
    e.health_max = e.health_max or e.health
    return e.events:listen('damage', function(info)
      e.sprite_color = {
        255,
        0,
        0
      }
      e.health = e.health - info.damage
      local cb
      cb = function()
        e.sprite_color = {
          255,
          255,
          255
        }
        e.ax = 0
        e.ay = 0
      end
      return state.modes.play:Timer(0.125, cb, 1)
    end)
  end
  sys.process = function(self, e, dt)
    if e.health <= 0 then
      return e.events:say("kill")
    end
  end
  return sys
end
local MeleeSystem
MeleeSystem = function(state)
  local sys = tiny.processingSystem()
  sys.filter = tiny.requireAll("tx", "ty", "melee_attack", "melee_range", "melee_damage")
  sys.onAdd = function(self, e)
    return e.events:listen("attack melee", function()
      local cx = e.tx + math.sin(e.rot) * (e.melee_range / 3)
      local cy = e.ty + math.cos(e.rot) * (e.melee_range / 3)
      local results = state.physics:queryRect(cx - (e.melee_range / 2), cy - (e.melee_range / 2), e.melee_range, e.melee_range, function(o)
        return o.health ~= nil
      end)
      for _index_0 = 1, #results do
        local _continue_0 = false
        repeat
          local o = results[_index_0]
          if o == e then
            _continue_0 = true
            break
          end
          local dx = o.tx - e.tx
          local dy = o.ty - e.ty
          dx, dy = Vector.normalize(dx, dy)
          o.ax = dx * o.acc
          o.ay = dy * o.acc
          o.events:say("damage", {
            damage = e.melee_damage,
            attacker = e
          })
          _continue_0 = true
        until true
        if not _continue_0 then
          break
        end
      end
    end)
  end
  return sys
end
local AnimSystem
AnimSystem = function(state)
  local sys = tiny.processingSystem()
  sys.filter = tiny.requireAll("tx", "ty", "anim_right", "anim_track")
  sys.onAdd = function(self, e)
    return e.events:listen("anim change", function(track, cb)
      if track ~= "walk" and track ~= "idle" then
        e.anim_running = true
      end
      e.anim_track = track
      e.anim = anim8.newAnimation(SPRITE_CACHE[e.sprite_image].grid(unpack(e.anim_tracks[e.anim_track])), 0.1, cb)
      if not e.anim_right then
        return e.anim:flipH()
      end
    end)
  end
  sys.process = function(self, e, dt)
    if e.rot >= 0 then
      if not e.anim_right then
        e.anim_right = true
        return e.anim:flipH()
      end
    else
      if e.anim_right then
        e.anim_right = false
        return e.anim:flipH()
      end
    end
  end
  return sys
end
local RenderSystem
RenderSystem = function(state)
  local sys = tiny.processingSystem()
  sys.filter = tiny.requireAll("tx", "ty", "rot", "sprite_image", "sprite_frame", "sprite_color")
  sys.onAdd = function(self, e)
    if SPRITE_CACHE[e.sprite_image] == nil then
      SPRITE_CACHE[e.sprite_image] = { }
      local image = g.newImage('assets/sprites/' .. e.sprite_image .. '.png')
      image:setFilter('nearest', 'nearest')
      SPRITE_CACHE[e.sprite_image].image = image
      SPRITE_CACHE[e.sprite_image].grid = anim8.newGrid(24, 24, image:getWidth(), image:getHeight())
    end
    e.anim = anim8.newAnimation(SPRITE_CACHE[e.sprite_image].grid(unpack(e.sprite_frame)), 0.1)
    e.sprite_ox = e.sprite_ox or 12
    e.sprite_oy = e.sprite_oy or 22
    e.sprite_scale = e.sprite_scale or 2
  end
  sys.process = function(self, e, dt)
    return e.anim:update(dt)
  end
  sys.draw = function(self)
    local t = { }
    local tkeys = { }
    local _list_0 = self.entities
    for _index_0 = 1, #_list_0 do
      local e = _list_0[_index_0]
      t[e.ty * 1000 + e.tx] = e
    end
    for k in pairs(t) do
      table.insert(tkeys, k)
    end
    table.sort(tkeys)
    for _, k in ipairs(tkeys) do
      self:draw_entity(t[k])
    end
  end
  sys.draw_entity = function(self, e)
    if e.active ~= false and state.map:is_visible(e.tx, e.ty) == 1 then
      g.push()
      g.translate(e.tx, e.ty)
      local base = 0
      if e.highlight ~= nil then
        g.setColor(255, 0, 0)
        e.anim:draw(SPRITE_CACHE[e.sprite_image].image, 2, base, 0, 2, 2, e.sprite_ox, e.sprite_oy)
        e.anim:draw(SPRITE_CACHE[e.sprite_image].image, -2, base, 0, 2, 2, e.sprite_ox, e.sprite_oy)
        e.anim:draw(SPRITE_CACHE[e.sprite_image].image, 0, base + 2, 0, 2, 2, e.sprite_ox, e.sprite_oy)
        e.anim:draw(SPRITE_CACHE[e.sprite_image].image, 0, base - 2, 0, 2, 2, e.sprite_ox, e.sprite_oy)
        g.setColor(255, 255, 255)
      end
      if e.hook then
        g.setColor(255, 0, 0, 255)
        local dx = e.tx - e.hook.tx
        local dy = e.ty - e.hook.ty
        g.line(0, base, -dx, -dy)
      end
      g.setColor(192, 192, 192, 255)
      g.setColor(unpack(e.sprite_color))
      e.anim:draw(SPRITE_CACHE[e.sprite_image].image, 0, base, 0, e.sprite_scale, e.sprite_scale, e.sprite_ox, e.sprite_oy)
      g.setColor(255, 255, 255)
      return g.pop()
    end
  end
  return sys
end
local MoveSystem
MoveSystem = function(state)
  local sys = tiny.processingSystem()
  sys.filter = tiny.requireAll("tx", "ty", "vx", "vy", "ax", "ay")
  sys._listeners = { }
  sys.onAdd = function(self, e)
    e.width, e.height = e.width or 6, e.height or 6
    e.physics_offsetx = e.physics_offsetx or (e.width / 2)
    e.physics_offsety = e.physics_offsety or (e.height / 2)
    e.drag = e.drag or 0.7
    e.speedMax = e.speedMax or 256
    e.vz = e.vz or 0
    e.tz = e.tz or 0
    e.speed = 0
    if not state.physics:hasItem(e) then
      state.physics:add(e, e.tx - e.physics_offsetx, e.ty - e.physics_offsety, e.width, e.height)
    else
      state.physics:update(e, e.tx - e.physics_offsetx, e.ty - e.physics_offsety)
    end
    sys._listeners[e] = e.events:listen("stop", function()
      e.vx = 0
      e.vy = 0
      e.ax = 0
      e.ay = 0
      e.speed = 0
    end)
  end
  sys.onRemove = function(self, e)
    e.events:stopListening(sys._listeners[e])
    sys._listeners[e] = nil
  end
  local filter
  filter = function(e, o)
    if e.solid == false or o.solid == false then
      return "cross"
    end
    return "slide"
  end
  local filterPit
  filterPit = function(e)
    if e.tile == 1 then
      return true
    end
    return false
  end
  sys.process = function(self, e, dt)
    if e.active == false then
      return 
    end
    if e.ax == 0 and e.tz <= 0 then
      e.vx = e.vx * e.drag
      if math.abs(e.vx) < 4 then
        e.vx = 0
      end
    else
      e.vx = e.vx + (e.ax * dt)
    end
    if e.ay == 0 and e.tz <= 0 then
      e.vy = e.vy * e.drag
      if math.abs(e.vy) < 4 then
        e.vy = 0
      end
    else
      e.vy = e.vy + (e.ay * dt)
    end
    e.speed = Vector.len(e.vx, e.vy)
    if e.speed > e.speedMax then
      e.vx, e.vy = Vector.normalize(e.vx, e.vy)
      e.vx = e.vx * e.speedMax
      e.vy = e.vy * e.speedMax
      e.speed = Vector.len(e.vx, e.vy)
    end
    if e.tz > 0 then
      e.vz = e.vz - 1
    end
    local dx = (e.vx * dt)
    local dy = (e.vy * dt)
    local dz = (e.vz * dt)
    if not state.physics:hasItem(e) then
      return 
    end
    local ax, ay, cols = state.physics:move(e, e.tx + dx - e.physics_offsetx, e.ty + dy - e.physics_offsety, filter)
    e.tx = ax + e.physics_offsetx
    e.ty = ay + e.physics_offsety
    if e.vx ~= 0 or e.vy ~= 0 then
      e.rot = math.atan2(e.vx, e.vy)
    end
    for _index_0 = 1, #cols do
      local c = cols[_index_0]
      c.item.events:say("collide", c.other)
      c.other.events:say("collide", c.item)
    end
    if dz ~= 0 or e.tz > 0 then
      e.tz = e.tz + dz
      if (e.tz < 0) then
        e.tz = 0
        e.vz = 0
        local results, len = state.physics:queryRect(e.tx, e.ty, e.width, e.height, filterPit)
        if #results > 0 then
          e.events:say("stop")
          e.events:say("stun")
          e.events:say("fall")
          state.physics:update(e, e.tx, e.ty)
        else
          e.events:say("land")
        end
      end
    end
    return e.events:say("move")
  end
  return sys
end
local PlayerSystem
PlayerSystem = function(state)
  local sys = tiny.processingSystem()
  sys.filter = tiny.requireAll("player", "acc", "ax", "ay")
  sys._listeners = { }
  sys.process = function(self, e, dt)
    e.ax = Controls.Horizontal() * e.acc
    e.ay = Controls.Vertical() * e.acc
    if e.anim_running ~= true then
      if e.vx ~= 0 or e.vy ~= 0 then
        if e.anim_track ~= "walk" then
          e.events:say("anim change", "walk")
        end
      else
        if e.anim_track ~= "idle" then
          e.events:say("anim change", "idle")
        end
      end
    end
    local interactions = state:search(e.tx, e.ty, 64, function(o)
      return o.interact ~= nil
    end)
    local d = 1e6
    local interact = nil
    for _index_0 = 1, #interactions do
      local i = interactions[_index_0]
      local dist = Vector.dist2(e.tx, e.ty, i.tx, i.ty)
      if dist < d then
        d = dist
        interact = i
      end
    end
    if interact ~= nil then
      interact.highlight = true
      e.interact_label = interact.interact.label
    else
      e.interact_label = nil
    end
    if Controls.Primary:pressed() then
      e.events:say("attack melee")
      e.events:say("anim change", "melee", function()
        e.anim_running = false
        return e.events:say("anim change", "idle")
      end)
    end
    if Controls.Secondary:pressed() then
      if e.ammo > 0 then
        e.ammo = e.ammo - 1
        e.events:say("attack ranged")
        e.events:say("anim change", "range", function()
          e.anim_running = false
          return e.events:say("anim change", "idle")
        end)
        d = {
          tx = e.tx + math.sin(e.rot) * 12,
          ty = e.ty - 10 + math.cos(e.rot) * 12,
          vx = math.sin(e.rot) * 2048,
          vy = math.cos(e.rot) * 2048,
          range_attacker = e
        }
        state:create_entity(d, 'bullet')
      end
    end
    if Controls.Tertiary:pressed() then
      if interact ~= nil then
        return interact.events:say("interact", e)
      end
    end
  end
  sys.onAdd = function(self, e)
    e.stun = false
    e.events:listen("stun", function()
      e.stun = true
    end)
    e.events:listen("move", function()
      state.cam:setPosition(math.floor(e.tx), math.floor(e.ty))
      return state.map:pov(math.floor(e.tx), math.floor(e.ty))
    end)
    e.events:listen("fall", function()
      return e.events:say("kill")
    end)
    e.events:listen("kill", function()
      local _list_0 = sys._listeners
      for _index_0 = 1, #_list_0 do
        local evt = _list_0[_index_0]
        state.events:stopListening(evt)
      end
      return state.events:say("player killed", e)
    end)
    return state.events:say("player created", e)
  end
  return sys
end
return {
  TriggerSystem = TriggerSystem,
  InteractSystem = InteractSystem,
  AnimSystem = AnimSystem,
  RenderSystem = RenderSystem,
  MoveSystem = MoveSystem,
  PlayerSystem = PlayerSystem,
  MeleeSystem = MeleeSystem,
  LivingSystem = LivingSystem,
  ScriptSystem = ScriptSystem
}
