table.shuffle = function(t)
  local rand = math.random
  assert(t, "shuffleTable() expected a table, got nil")
  local iterations = #t
  local j = nil
  for i = iterations, 2, -1 do
    j = rand(i)
    t[i], t[j] = t[j], t[i]
  end
end
math.sign = function(n)
  if n > 0 then
    return 1
  end
  if n < 0 then
    return -1
  end
  return 0
end
local remove_by_val
remove_by_val = function(list, val)
  for idx = #list, 1, -1 do
    if list[idx] == val then
      table.remove(list, idx)
    end
  end
end
local pprint
pprint = function(obj, indent)
  if indent == nil then
    indent = 0
  end
  if type(obj) ~= 'table' then
    return print(obj)
  else
    for k, v in pairs(obj) do
      if type(v) == 'table' then
        print(string.format("%s%s:", string.rep(" ", indent), k))
        pprint(v, indent + 4)
      else
        print(string.format("%s%s: %s", string.rep(" ", indent), k, v))
      end
    end
  end
end
local deep_copy
deep_copy = function(obj, seen)
  if type(obj) ~= 'table' then
    return obj
  end
  if seen and seen[obj] then
    return seen[obj]
  end
  local s = seen or { }
  local res = setmetatable({ }, getmetatable(obj))
  s[obj] = res
  for k, v in pairs(obj) do
    res[deep_copy(k, s)] = deep_copy(v, s)
  end
  return res
end
local deep_extend
deep_extend = function(obj, update)
  for k, v in pairs(update) do
    if type(v) == 'table' then
      if obj[k] == nil then
        obj[k] = deep_copy(v)
      else
        deep_extend(obj[k], v)
      end
    else
      obj[k] = v
    end
  end
end
return {
  remove_by_val = remove_by_val,
  pprint = pprint,
  deep_copy = deep_copy,
  deep_extend = deep_extend
}
